@@ -1,11 +1,14 @@
 """Tools for performing image registration."""
+import matplotlib.pyplot as plt
+import numpy as np
 import os
 import shutil
 import subprocess
 
-from skrt.image import Image
+import skrt.image
 from skrt.structures import ROI, StructureSet
-from skrt.core import Data
+from skrt.core import get_logger, Data, to_list, Defaults
+from skrt.dose import ImageOverlay
 
 _ELASTIX_DIR = None
 _ELASTIX = "elastix"
@@ -13,9 +16,10 @@ _transformix = "transformix"
 
 
 class Registration(Data):
+
     def __init__(
-        self, path, fixed=None, moving=None, pfiles=None, auto=False, overwrite=False
-    ):
+        self, path, fixed=None, moving=None, pfiles=None, auto=False,
+        overwrite=False, capture_output=False, log_level=None):
         """Load data for an image registration and run the registration if
         auto_seg=True.
 
@@ -59,8 +63,25 @@ class Registration(Data):
         overwrite : bool, default=False
             If True and <path> already contains files, these will be deleted,
             meaning that no prior registration results will be loaded.
+
+        capture_output : bool, default=False
+            If True, capture to stdout messages from performing
+            registration and transformation.  This suppresses
+            informational messages to screen, but doesn't suppress error
+            messages.
+
+        log_level: str/int/None, default=None
+            Severity level for event logging.  If the value is None,
+            log_level is set to the value of skrt.core.Defaults().log_level.
         """
 
+        # Set up event logging and output capture
+        self.log_level = \
+                Defaults().log_level if log_level is None else log_level
+        self.logger = get_logger(
+                name=type(self).__name__, log_level=self.log_level)
+        self.capture_output = capture_output
+
         # Set up directory
         self.path = path
         if not os.path.exists(path):
@@ -86,6 +107,8 @@ class Registration(Data):
         self.pfiles = {}
         self.tfiles = {}
         self.transformed_images = {}
+        self.jacobians = {}
+        self.deformation_fields = {}
         if isinstance(pfiles, str):
             pfiles = [pfiles]
         if pfiles is not None:
@@ -120,12 +143,12 @@ class Registration(Data):
                 "should be either fixed or moving"
             )
 
-        if not isinstance(im, Image):
-            im = Image(im)
+        if not isinstance(im, skrt.image.Image):
+            im = skrt.image.Image(im)
         path = getattr(self, f"{category}_path")
         if not os.path.exists(path) or force:
-            Image.write(im, path)
-        setattr(self, f"{category}_image", Image(path))
+            skrt.image.Image.write(im, path)
+        setattr(self, f"{category}_image", skrt.image.Image(path))
 
     def set_fixed_image(self, im):
         """Assign a fixed image."""
@@ -218,7 +241,7 @@ class Registration(Data):
                 p = pfiles[p]
             else:
                 name = None
-            self.add_pfile(p)
+            self.add_pfile(p, name=name)
 
         self.write_steps()
 
@@ -229,7 +252,7 @@ class Registration(Data):
         informative purposes.
         """
 
-        for file in get_default_params():
+        for file in get_default_pfiles():
             print(file)
 
     def get_default_params(self, filename):
@@ -337,10 +360,26 @@ class Registration(Data):
             # Check for transformed moving images
             im_path = os.path.join(outdir, "result.0.nii")
             if os.path.exists(im_path):
-                self.transformed_images[step] = Image(
+                self.transformed_images[step] = skrt.image.Image(
                     im_path, title="Transformed moving"
                 )
 
+            # Check for Jacobian determinant
+            jac_path = os.path.join(outdir, "spatialJacobian.nii")
+            if os.path.exists(jac_path):
+                self.jacobians[step] = Jacobian(
+                    jac_path, title="Jacobian determinant",
+                    image=self.transformed_images[step]
+                )
+
+            # Check for deformation field
+            df_path = os.path.join(outdir, "deformationField.nii")
+            if os.path.exists(df_path):
+                self.deformation_fields[step] = DeformationField(
+                    df_path, title="Deformation field",
+                    image=self.transformed_images[step]
+                )
+
     def clear(self):
         """Remove all registration steps and their output directories."""
 
@@ -365,11 +404,8 @@ class Registration(Data):
         """Get elastix registration command for a given step."""
 
         # Get step number
-        if isinstance(step, int):
-            i = step
-            step = self.steps[step]
-        else:
-            i = self.steps.index(step)
+        i = self.get_step_number(step)
+        step = self.get_step_name(step)
 
         # Check for input transform file
         tfile = None
@@ -433,13 +469,7 @@ class Registration(Data):
         # Check all steps exist and convert numbers to names
         steps = []
         for step in steps_input:
-            if isinstance(step, int):
-                if step >= len(self.steps):
-                    raise RuntimeError(f"Step {step} not found!")
-                step = self.steps[step]
-            if step not in self.steps:
-                raise RuntimeError(f"Step {step} not found!")
-            steps.append(step)
+            steps.append(self.get_step_name(step))
 
         # Run registration for each step
         for step in steps:
@@ -471,23 +501,21 @@ class Registration(Data):
         """
 
         # Check if the registration has already been performed
-        if self.already_performed(step) and not force:
+        if self.is_registered(step) and not force:
             return
 
         # Check that previous step has been run if needed
-        if isinstance(step, int):
-            i = step
-            step = self.steps[i]
-        else:
-            i = self.steps.index(step)
+        i = self.get_step_number(step)
+        step = self.get_step_name(step)
         if use_previous_tfile and i > 0:
-            if not self.already_performed(i - 1):
+            if not self.is_registered(i - 1):
                 self.register(i, use_previous_tfile=True)
 
         # Run
         cmd = self.get_ELASTIX_cmd(step, use_previous_tfile)
-        print("Running command:\n", cmd)
-        code = subprocess.run(cmd.split()).returncode
+        self.logger.info(f"Running command:\n {cmd}")
+        code = subprocess.run(
+                cmd.split(), capture_output=self.capture_output).returncode
 
         # Check whether registration succeeded
         if code:
@@ -501,26 +529,23 @@ class Registration(Data):
             self.tfiles[step] = os.path.join(
                 self.outdirs[step], "TransformParameters.0.txt"
             )
-            self.transformed_images[step] = Image(
+            self.transformed_images[step] = skrt.image.Image(
                 os.path.join(self.outdirs[step], "result.0.nii"),
                 title="Transformed moving",
             )
 
-    def already_performed(self, step):
-        """Check whether a registration step has already performed (i.e. has
-        a valid output transform file)."""
+    def is_registered(self, step):
+        """Check whether a registration step has already been performed (i.e. 
+        has a valid output transform file)."""
 
-        if isinstance(step, int):
-            step = self.steps[step]
+        step = self.get_step_name(step)
         return step in self.tfiles and os.path.exists(self.tfiles[step])
 
     def print_log(self, step=-1):
         """Print elastix output log for a given step (by default, the 
         last step)."""
 
-        if isinstance(step, int):
-            step = self.steps[step]
-
+        step = self.get_step_name(step)
         logfile = os.path.join(self.outdirs[step], "elastix.log")
         if not os.path.exists(logfile):
             print(f"No log found - try running registration step {step}.")
@@ -536,19 +561,19 @@ class Registration(Data):
         `to_transform`. Functions called depending on type are:
 
         Image:
-            transform_image(to_transform, \*\*kwargs)
+            transform_image(to_transform, `**`kwargs)
         str:
-            transform_nifti(to_transform, \*\*kwargs)
+            transform_data(to_transform, `**`kwargs)
         StructureSet:
-            transform_structure_set(to_transform, \*\*kwargs)
+            transform_structure_set(to_transform, `**`kwargs)
         ROI:
-            transform_roi(to_transform, \*\*kwargs)
+            transform_roi(to_transform, `**`kwargs)
         """
 
-        if issubclass(type(to_transform), Image):
+        if issubclass(type(to_transform), skrt.image.Image):
             return self.transform_image(to_transform, **kwargs)
         elif isinstance(to_transform, str):
-            return self.transform_nifti(to_transform, **kwargs)
+            return self.transform_data(to_transform, **kwargs)
         elif isinstance(to_transform, ROI):
             return self.transform_roi(to_transform, **kwargs)
         elif isinstance(to_transform, StructureSet):
@@ -556,7 +581,8 @@ class Registration(Data):
         else:
             print(f"Unrecognised transform input type {type(to_transform)}")
 
-    def transform_image(self, im, step=-1, outfile=None, params=None, rois=None):
+    def transform_image(self, im, step=-1, outfile=None, params=None, rois=None,
+            ):
         """
         Transform an image using the output transform from a given
         registration step (by default, the final step). If the registration
@@ -587,7 +613,7 @@ class Registration(Data):
         """
 
         # Save image temporarily as nifti if needed
-        im = Image(im)
+        im = skrt.image.Image(im)
         self.make_tmp_dir()
         if im.source_type == "nifti":
             im_path = im.path
@@ -596,7 +622,7 @@ class Registration(Data):
             im.write(im_path, verbose=False)
 
         # Transform the nifti file
-        result_path = self.transform_nifti(im_path, step, params)
+        result_path = self.transform_data(im_path, step, params)
         if result_path is None:
             return
 
@@ -606,7 +632,7 @@ class Registration(Data):
             return
 
         # Otherwise, return Image object
-        final_im = Image(result_path)
+        final_im = skrt.image.Image(result_path)
         self.rm_tmp_dir()
 
         # Transform structure sets
@@ -622,20 +648,15 @@ class Registration(Data):
 
         return final_im
 
-    def transform_nifti(self, path, step=-1, params=None):
-        """Transform a nifti file at a given path for a given step, ensuring
-        that the step has been run. Return the path to the transformed file
-        inside self._tmp_dir."""
-
-        if isinstance(step, int):
-            i = step
-            step = self.steps[i]
-        else:
-            i = self.steps.index(step)
+    def transform_data(self, path, step=-1, params=None):
+        """Transform a nifti file or point cloud at a given path
+        for a given step, ensuring that the step has been run. Return
+        the path to the transformed file inside self._tmp_dir."""
 
         # Check registration has been performed, and run it if not
-        if not self.already_performed(step):
-            self.register(self.steps[: i + 1])
+        i = self.get_step_number(step)
+        step = self.get_step_name(step)
+        self.ensure_registered(step)
 
         # Make temporary modified parameter file if needed
         self.make_tmp_dir()
@@ -645,18 +666,27 @@ class Registration(Data):
             tfile = os.path.join(self._tmp_dir, "TransformParameters.txt")
             adjust_parameters(self.tfiles[step], tfile, params)
 
+        # Define parameters specific to data type.
+        if '.txt' == os.path.splitext(path)[1]:
+            option = '-def'
+            outfile = 'outputpoints.txt' 
+        else:
+            option = '-in'
+            outfile = 'result.nii'
+
         # Run transformix
         cmd = [
             _transformix,
-            "-in",
+            option,
             path.replace("\\", "/"),
             "-out",
             self._tmp_dir,
             "-tp",
             tfile,
         ]
-        print("running command:", " ".join(cmd))
-        code = subprocess.run(cmd).returncode
+        self.logger.info(f'Running command:\n {" ".join(cmd)}')
+        code = subprocess.run(
+                cmd, capture_output=self.capture_output).returncode
 
         # If command failed, move log out from temporary dir
         if code:
@@ -671,9 +701,10 @@ class Registration(Data):
             return
 
         # Return path to result
-        return os.path.join(self._tmp_dir, "result.nii")
+        return os.path.join(self._tmp_dir, outfile)
 
-    def transform_roi(self, roi, step=-1, outfile=None, params=None):
+    def transform_roi(self, roi, step=-1, outfile=None, params=None,
+            transform_points=False):
         """Transform a single ROI using the output transform from a given
         registration step (by default, the final step). If the registration
         step has not yet been performed, the step and all preceding steps
@@ -701,16 +732,25 @@ class Registration(Data):
             so you will need to use double quotes.
             By default, "ResampleInterpolator" will be set to
             "FinalNearestNeighborInterpolator".
+
+        transform_points : bool, default=False
+           If False, the transform is applied to pull the ROI mask
+           from the reference frame of the moving image to the reference
+           frame of the fixed image.  If True, the transform is applied
+           to push ROI contour points from the reference frame of the
+           fixed image to the reference frame of the moving image.
         """
 
         # Save ROI temporarily as nifti if needed
         roi = ROI(roi)
         roi.load()
         self.make_tmp_dir()
-        if roi.source_type == "mask" and roi.mask.source_type == "nifti":
+        if (roi.source_type == "mask" and roi.mask.source_type == "nifti"
+                and not transform_points):
             roi_path = roi.mask.path
         else:
-            roi_path = os.path.join(self._tmp_dir, f"{roi.name}.nii.gz")
+            ext = 'txt' if transform_points else 'nii.gz'
+            roi_path = os.path.join(self._tmp_dir, f"{roi.name}.{ext}")
             roi.write(roi_path, verbose=False)
 
         # Set default parameters
@@ -718,8 +758,8 @@ class Registration(Data):
         if params is not None:
             default_params.update(params)
 
-        # Transform the nifti file
-        result_path = self.transform_nifti(roi_path, step, default_params)
+        # Transform the nifti file or point cloud
+        result_path = self.transform_data(roi_path, step, default_params)
         if result_path is None:
             return result_path
 
@@ -731,12 +771,15 @@ class Registration(Data):
         # Otherwise, return ROI object
         roi = ROI(result_path, name=roi.name, color=roi.color)
         self.rm_tmp_dir()
-        roi.set_image(self.get_transformed_image(step))
+        if transform_points:
+            roi.set_image(self.moving_image)
+        else:
+            roi.set_image(self.get_transformed_image(step))
         return roi
 
     def transform_structure_set(
-        self, structure_set, step=-1, outfile=None, params=None
-    ):
+        self, structure_set, step=-1, outfile=None, params=None,
+        transform_points=False):
         """Transform a structure set using the output transform from a given
         registration step (by default, the final step). If the registration
         step has not yet been performed, the step and all preceding steps
@@ -764,34 +807,44 @@ class Registration(Data):
             so you will need to use double quotes.
             By default, "ResampleInterpolator" will be set to
             "FinalNearestNeighborInterpolator".
+
+        transform_points : bool, default=False
+           If False, the transform is applied to pull ROI masks
+           from the reference frame of the moving image to the reference
+           frame of the fixed image.  If True, the transform is applied
+           to push ROI contour points from the reference frame of the
+           fixed image to the reference frame of the moving image.
         """
 
         final = StructureSet()
         for roi in structure_set:
-            transformed_roi = self.transform_roi(roi, step, params=params)
+            transformed_roi = self.transform_roi(roi, step, params=params,
+                    transform_points=transform_points)
             if transformed_roi is not None:
                 final.add_roi(transformed_roi)
 
         # Write structure set if outname is given
         if outfile is not None:
-            final.write(outname)
+            final.write(outfile)
             return
 
         # Otherwise, return structure set
         final.name = "Transformed"
-        final.set_image(self.get_transformed_image(step))
+        if transform_points:
+            final.set_image(self.moving_image)
+        else:
+            final.set_image(self.get_transformed_image(step))
         return final
 
     def transform_moving_image(self, step=-1):
         """Transform the moving image using the output of a registration step
         and set it to self.transformed_images[step]."""
 
-        if isinstance(step, int):
-            step = self.steps[step]
-
+        step = self.get_step_name(step)
         outfile = os.path.join(self.outdirs[step], "result.0.nii")
         self.transform(self.moving_image, outfile=outfile)
-        self.transformed_images[step] = Image(outfile, title="Transformed moving")
+        self.transformed_images[step] = skrt.image.Image(
+            outfile, title="Transformed moving")
 
     def get_transformed_image(self, step=-1, force=False):
         """Get the transformed moving image for a given step, by default the
@@ -799,20 +852,29 @@ class Registration(Data):
         transformix even if there is already a resultant image in the
         output directory for that step."""
 
-        if isinstance(step, int):
-            step = self.steps[step]
-
         # Run registration if needed
-        if not self.already_performed(step):
-            self.register(step)
+        step = self.get_step_name(step)
+        was_registered = self.ensure_registered(step)
 
-        # Otherwise if forcing, re-transform the moving image
-        elif force:
+        # If forcing and registration had already been done, re-transform the 
+        # moving image (otherwise, moving image will have just been recreated
+        # anyway by running registration)
+        if force and was_registered:
             self.transform_moving_image(step)
 
         # Return clone of the transformed image object
         return self.transformed_images[step].clone()
 
+    def ensure_registered(self, step):
+        """If a step has not already been registered, perform registration
+        for this step and any preceding steps. Return True if registration
+        had already been performed."""
+
+        if not self.is_registered(step):
+            self.register_step(step)
+            return False
+        return True
+
     def make_tmp_dir(self):
         """Create temporary directory."""
 
@@ -829,12 +891,26 @@ class Registration(Data):
             shutil.rmtree(self._tmp_dir)
 
     def adjust_pfile(self, step, params, reset=True):
-        """Adjust parameters in an input parameter file for a given step. If
-        reset=True, this will force the registration to be re-run for that step.
+        """Adjust parameters in an input parameter file for a given step. 
+
+        **Parameters:**
+
+        step : str/int
+            Name or number of step for which the parameter file should be 
+            adjusted.
+
+        params : dict
+            Dict of parameter names and desired parameter values. Any parameters
+            matching those in the existing file will be overwritten with the
+            new values. Any parameters not in the existing file will be added.
+            Parameters in the existing file but not in <params> will remain
+            unchanged.
+
+        reset : bool, default=True
+            If True, this will remove existing registration results.
         """
 
-        if isinstance(step, int):
-            step = self.steps[step]
+        step = self.get_step_name(step)
         adjust_parameters(self.pfiles[step], self.pfiles[step], params)
         if reset and step in self.tfiles:
             del self.tfiles[step]
@@ -844,7 +920,10 @@ class Registration(Data):
 
         from skrt.better_viewer import BetterViewer
 
-        kwargs.setdefault("hu", self.fixed_image.default_window)
+        kwargs.setdefault(
+            "intensity", 
+            [self.fixed_image._default_vmin, self.fixed_image._default_vmax]
+        )
         kwargs.setdefault("match_axes", "y")
         kwargs.setdefault("title", ["Fixed", "Moving"])
         kwargs.setdefault("comparison", True)
@@ -865,20 +944,22 @@ class Registration(Data):
             If True, the result will be displayed in comparison with the
             fixed image.
 
-        \*\*kwargs :
+        `**`kwargs :
             Optional keyword arguments to pass to BetterViewer.
         """
 
         from skrt.better_viewer import BetterViewer
 
-        if isinstance(step, int):
-            step = self.steps[step]
+        step = self.get_step_name(step)
         if step not in self.transformed_images:
-            self.register(step)
+            self.register_step(step)
         if compare_with_fixed:
             ims = [self.fixed_image, self.transformed_images[step]]
             kwargs.setdefault("comparison", True)
-            kwargs.setdefault("hu", self.fixed_image.default_window)
+            kwargs.setdefault(
+                "intensity", 
+                [self.fixed_image._default_vmin, self.fixed_image._default_vmax]
+            )
             kwargs.setdefault("title", ["Fixed", "Transformed moving"])
         else:
             ims = self.transformed_images[step]
@@ -907,9 +988,7 @@ class Registration(Data):
         """
 
         # Get step name
-        if isinstance(step, int):
-            step = self.steps[step]
-        elif step is None:
+        if step is None:
             if len(self.steps) == 1:
                 step = self.steps[0]
             else:
@@ -919,9 +998,10 @@ class Registration(Data):
                     "Registration.manually_adjust_transform()."
                 )
                 return
+        step = self.get_step_name(step)
 
         # Check registration has been run
-        if not self.already_performed(step):
+        if not self.is_registered(step):
             print(f"Registration for {step} has not yet been performed.")
             return
 
@@ -956,8 +1036,7 @@ class Registration(Data):
         Get dict of input parameters for a given step.
         """
 
-        if isinstance(step, int):
-            step = self.steps[step]
+        step = self.get_step_name(step)
         return read_parameters(self.pfiles[step])
 
     def get_transform_parameters(self, step):
@@ -965,14 +1044,324 @@ class Registration(Data):
         Get dict of output transform parameters for a given step.
         """
 
-        if isinstance(step, int):
-            step = self.steps[step]
-
-        if not self.already_performed(step):
+        step = self.get_step_name(step)
+        if not self.is_registered(step):
             print(f"Registration step {step} has not yet been performed.")
             return
         return read_parameters(self.tfiles[step])
 
+    def get_step_name(self, step):
+        """Convert <step> to a string containing a step name. If <step> is
+        already a string, check it's a valid step and return it; otherwise
+        if <step> is an integer, return the corresponding step name."""
+
+        if isinstance(step, str):
+            if step not in self.steps:
+                raise RuntimeError(f"Step {step} not a valid registration step")
+            return step
+        
+        return self.steps[step]
+
+    def get_step_number(self, step):
+        """Convert <step> to an int containing the number of a given step.
+        If <step> is an int, check it corresponds to a step number, ensure it 
+        is positive, and return it. Otherwise if it's a string, return the 
+        index of that step name in self.steps."""
+
+        if isinstance(step, int):
+            if step >= len(self.steps) or step < -len(self.steps):
+                raise IndexError(f"Step {step} not a valid index number for "
+                                 f"step list of length {len(self.steps)}")
+            if step < 0:
+                return len(self.steps) + step
+            return step
+
+        try:
+            return self.steps.index(step)
+        except ValueError:
+            raise ValueError(f"Step {step} not found")
+
+    def _get_jac_or_def(self, step, is_jac, force):
+
+        # Settings for jacobian or deformation field
+        if is_jac:
+            storage = self.jacobians
+        else:
+            storage = self.deformation_fields
+
+        # If object already exists, return it unless forcing
+        step = self.get_step_name(step)
+        if step in storage and not force:
+            return storage[step]
+
+        # Ensure registration has been performed for this step
+        self.ensure_registered(step)
+
+        # Create new object
+        storage[step] = run_transformix_on_all(
+            is_jac, outdir=self.outdirs[step], tfile=self.tfiles[step], 
+            image=self.transformed_images[step]
+        )
+        return storage[step]
+
+    def get_jacobian(self, step=-1):
+        """Generate Jacobian determinant using transformix for a given 
+        registration step (or return existing Jacobian object, unless 
+        force=True).
+        """
+
+        return self._get_jac_or_def(step, True, force)
+
+    def get_deformation_field(self, step=-1, force=False):
+        """Generate deformation field using transformix for a given 
+        registration step."""
+        
+        return self._get_jac_or_def(step, False, force)
+
+
+class Jacobian(ImageOverlay):
+
+    def __init__(self, *args, **kwargs):
+
+        ImageOverlay.__init__(self, *args, **kwargs)
+
+        # Plot settings specific to Jacobian determinant
+        self._default_cmap = "seismic"
+        self._default_colorbar_label = "Jacobian"
+        self._default_vmin = -0.5
+        self._default_vmax = 2.5
+
+    def view(self, **kwargs):
+        return ImageOverlay.view(self, kwarg_name="jacobian", **kwargs)
+
+
+class DeformationField:
+    """Class representing a vector field."""
+
+    def __init__(self, path, image=None, **kwargs):
+        """Load vector field."""
+
+        # Initialise own image object
+        self._image = skrt.image.Image(path, **kwargs)
+
+        # Assign an associated Image
+        self.image = image
+
+    def load(self, force=False):
+
+        self._image.load()
+        assert self._image.get_data().ndim == 4
+
+    def get_slice(self, view, sl=None, idx=None, pos=None, scale_in_mm=True):
+        """Get voxel positions and displacement vectors on a 2D slice."""
+
+        data = self._image.get_slice(view, sl=sl, idx=idx, pos=pos)
+        x_ax, y_ax = skrt.image._plot_axes[view]
+
+        # Get x/y displacement vectors
+        df_x = np.squeeze(data[:, :, x_ax])
+        df_y = np.squeeze(data[:, :, y_ax])
+        if not scale_in_mm:
+            df_x /= self._image.voxel_size[x_ax]
+            df_y /= self._image.voxel_size[y_ax]
+
+        # Get x/y coordinates of each point on the slice
+        xs = np.arange(0, data.shape[1])
+        ys = np.arange(0, data.shape[0])
+        if scale_in_mm:
+            xs = self._image.origin[x_ax] + xs * self._image.voxel_size[x_ax]
+            ys = self._image.origin[y_ax] + ys * self._image.voxel_size[y_ax]
+        y, x = np.meshgrid(ys, xs)
+        x = x.T
+        y = y.T
+        return x, y, df_x, df_y
+
+    def plot(
+        self,
+        view="x-y",
+        sl=None,
+        idx=None,
+        pos=None,
+        plot_type="quiver",
+        include_image=False,
+        spacing=30,
+        ax=None,
+        gs=None,
+        figsize=None,
+        zoom=None,
+        zoom_centre=None,
+        show=True,
+        save_as=None,
+        scale_in_mm=True,
+        **kwargs
+    ):
+
+        # Set up axes
+        self._image.set_ax(view, ax, gs, figsize, zoom)
+        self.ax = self._image.ax
+        self.fig = self._image.fig
+
+        # Plot the underlying image
+        if include_image and self.image is not None:
+            self.image.plot(view, ax=self.ax, show=False)
+
+        # Get spacing in each direction in number of voxels
+        spacing = self.convert_spacing(spacing, scale_in_mm)
+
+        # Get vectors and positions on this slice
+        data_slice = self.get_slice(view, sl=sl, idx=idx, pos=pos, 
+                                    scale_in_mm=scale_in_mm)
+
+        # Create plot
+        if plot_type == "quiver":
+            self._plot_quiver(view, data_slice, spacing, **kwargs)
+        elif plot_type == "grid":
+            self._plot_grid(view, data_slice, spacing, **kwargs)
+        else:
+            raise ValueError(f"Unrecognised plot type {plot_type}")
+
+        # Label and zoom axes
+        idx = self._image.get_idx(view, sl=sl, idx=idx, pos=pos)
+        self._image.label_ax(view, idx=idx, scale_in_mm=scale_in_mm, **kwargs)
+        self._image.zoom_ax(view, zoom, zoom_centre)
+
+        # Display image
+        plt.tight_layout()
+        if show:
+            plt.show()
+
+        # Save to file
+        if save_as:
+            self.fig.savefig(save_as)
+            plt.close()
+
+    def _plot_quiver(
+        self, 
+        view, 
+        data_slice,
+        spacing,
+        mpl_kwargs=None, 
+    ):
+        """Draw a quiver plot."""
+
+        # Get arrow positions and lengths
+        x_ax, y_ax = skrt.image._plot_axes[view]
+        x, y, df_x, df_y = data_slice
+        arrows_x = df_x[:: spacing[y_ax], :: spacing[x_ax]]
+        arrows_y = -df_y[:: spacing[y_ax], :: spacing[x_ax]]
+        plot_x = x[:: spacing[y_ax], :: spacing[x_ax]]
+        plot_y = y[:: spacing[y_ax], :: spacing[x_ax]]
+
+        # Make plotting kwargs
+        default_kwargs = {"cmap": "jet"}
+        if mpl_kwargs is not None:
+            default_kwargs.update(mpl_kwargs)
+
+        # Plot arrows
+        if arrows_x.any() or arrows_y.any():
+            M = np.hypot(arrows_x, arrows_y)
+            self.ax.quiver(
+                plot_x,
+                plot_y,
+                arrows_x,
+                arrows_y,
+                M,
+                **default_kwargs
+            )
+        else:
+            # If arrow lengths are zero, plot dots
+            ax.scatter(plot_x, plot_y, c="navy", marker=".")
+
+    def _plot_grid(
+        self, 
+        view, 
+        data_slice,
+        spacing,
+        mpl_kwargs=None, 
+    ):
+        """Draw a grid plot."""
+
+        # Get gridline positions
+        #  self.ax.autoscale(False)
+        x_ax, y_ax = skrt.image._plot_axes[view]
+        x, y, df_x, df_y = data_slice
+        grid_x = x + df_x
+        grid_y = y + df_y
+
+        # Make plotting kwargs
+        default_kwargs = {"color": "green", "linewidth": 2}
+        if mpl_kwargs is not None:
+            default_kwargs.update(mpl_kwargs)
+
+        # Plot gridlines
+        for i in np.arange(0, x.shape[0], spacing[y_ax]):
+            self.ax.plot(grid_x[i, :], grid_y[i, :], **default_kwargs)
+        for j in np.arange(0, x.shape[1], spacing[x_ax]):
+            self.ax.plot(grid_x[:, j], grid_y[:, j], **default_kwargs)
+
+    def convert_spacing(self, spacing, scale_in_mm):
+        """Convert grid spacing in mm or voxels to list containing grid spacing 
+        in voxels in each dimension."""
+
+        self.load()
+
+        spacing = to_list(spacing)
+        output = []
+        if scale_in_mm:
+            output = [abs(round(spacing[i] / self._image.voxel_size[i])) 
+                      for i in range(3)]
+        else:
+            output = spacing
+
+        # Ensure spacing is at least 2 voxels
+        for i, sp in enumerate(output):
+            if sp < 2:
+                output[i] = 2
+
+        return output
+
+
+
+def run_transformix_on_all(is_jac, outdir, tfile, image=None):
+    """Run transformix with either `-jac all` or `-def all` to create a
+    Jacobian determinant or deformation field file, and return either
+    a Jacobian or DeformationField object initialised from the output file.
+    """
+
+    # Settings
+    if is_jac:
+        opt = "-jac"
+        dtype = Jacobian
+        expected_outname = "spatialJacobian.nii"
+        name = "Jacobian determinant"
+    else:
+        opt = "-def"
+        dtype = DeformationField
+        expected_outname = "deformationField.nii"
+        name = "Deformation field"
+
+    # Run transformix
+    cmd = [
+        _transformix,
+        opt,
+        "all",
+        "-out",
+        outdir,
+        "-tp",
+        tfile
+    ]
+    self.logger.info('Running command:\n {" ".join(cmd)}')
+    code = subprocess.run(cmd, capture_output=self.capture_output).returncode
+    if code:
+        logfile = os.path.join(outdir, 'transformix.log')
+        raise RuntimeError(f"Jacobian creation failed. See {logfile} for "
+                           " more info.")
+
+    # Create output object
+    output_file = os.path.join(outdir, expected_outname)
+    assert os.path.exists(output_file)
+    return dtype(output_file, image=image, name=name)
+
 
 def set_elastix_dir(path):
 
@@ -986,13 +1375,46 @@ def set_elastix_dir(path):
     if os.path.exists(os.path.join(_ELASTIX_DIR, "bin/elastix")):
         _ELASTIX = os.path.join(_ELASTIX_DIR, "bin/elastix")
         _transformix = os.path.join(_ELASTIX_DIR, "bin/transformix")
+        lib_dir = os.path.join(_ELASTIX_DIR, 'lib')
     elif os.path.exists(os.path.join(_ELASTIX_DIR, "elastix.exe")):
         _ELASTIX = os.path.join(_ELASTIX_DIR, "elastix.exe")
         _transformix = os.path.join(_ELASTIX_DIR, "transformix.exe")
+        lib_dir = os.path.join(os.path.dirname(_ELASTIX_DIR), 'lib')
     else:
         raise RuntimeError(f"No elastix executable found in {_ELASTIX_DIR}!")
 
 
+    # Cover Linux and MacOS
+    prepend_path('DYLD_FALLBACK_LIBRARY_PATH', lib_dir)
+    prepend_path('LD_LIBRARY_PATH', lib_dir)
+
+def prepend_path(variable, path, path_must_exist=True):
+    '''
+    Prepend path to environment variable.
+
+    **Parameters:**
+
+    variable : str
+        Environment variable to which path is to be prepended.
+
+    path : str
+        Path to be prepended.
+
+    path_must_exist : bool, default=True
+        If True, only append path if it exists.
+    '''
+    path_ok = True
+    if path_must_exist:
+        if not os.path.exists(path):
+            path_ok = False
+
+    if path_ok:
+        if variable in os.environ:
+            if os.environ[variable]:
+                os.environ[variable] = f'{path}:{os.environ[variable]}'
+        else:
+            os.environ[variable] = path
+
 def adjust_parameters(infile, outfile, params):
     """Open an elastix parameter file (works for both input parameter and
     output transform files), adjust its parameters, and save it to a new
